import argparse
import os
import sys

from wdb import Wdb
from wdb._compat import execute

parser = argparse.ArgumentParser(description='Wdb, the web python debugger.')
parser.add_argument(
    '--source',
    dest='source',
    help='Source the specified file before openning the shell',
)

parser.add_argument(
    '--trace',
    dest='trace',
    action='store_true',
    help='Activate trace (otherwise just inspect tracebacks).',
)
parser.add_argument('file', nargs='?', help='the path to the file to debug.')
parser.add_argument('args', nargs='*', help='arguments to the debugged file.')


def main():
    """Wdb entry point"""
    sys.path.insert(0, os.getcwd())
    args, extrargs = parser.parse_known_args()
    sys.argv = ['wdb'] + args.args + extrargs

    if args.file:
        file = os.path.join(os.getcwd(), args.file)
        if args.source:
            print('The source argument cannot be used with file.')
            sys.exit(1)

        if not os.path.exists(file):
            print('Error:', file, 'does not exist')
            sys.exit(1)
        if args.trace:
            Wdb.get().run_file(file)
        else:

            def wdb_pm(xtype, value, traceback):
                sys.__excepthook__(xtype, value, traceback)
                wdb = Wdb.get()
                wdb.reset()
                wdb.interaction(None, traceback, post_mortem=True)

            sys.excepthook = wdb_pm

            with open(file) as f:
                code = compile(f.read(), file, 'exec')
                execute(code, globals(), globals())

    else:
        source = None
        if args.source:
            source = os.path.join(os.getcwd(), args.source)
            if not os.path.exists(source):
                print('Error:', source, 'does not exist')
                sys.exit(1)

        Wdb.get().shell(source)


if __name__ == '__main__':
    main()



# *-* coding: utf-8 *-*
import os
import re
import sys
import time
import token as tokens
import traceback
from base64 import b64encode
from logging import WARNING
from subprocess import Popen
from tokenize import TokenError, generate_tokens

from . import __version__, _initial_globals
from ._compat import (
    JSONEncoder,
    StringIO,
    _detect_lines_encoding,
    dumps,
    escape,
    execute,
    force_bytes,
    from_bytes,
    is_str,
    loads,
    logger,
    quote,
    u,
)
from .utils import (
    Html5Diff,
    executable_line,
    get_doc,
    get_source,
    importable_module,
    inplace,
    search_key_in_obj,
    search_value_in_obj,
    timeout_of,
)

try:
    from cutter import cut
    from cutter.utils import bang_compile as compile
except ImportError:
    cut = None

try:
    import magic
except ImportError:
    magic = None

try:
    from jedi import Interpreter
except ImportError:
    Interpreter = None

log = logger('wdb.ui')


def eval_(src, *args, **kwargs):
    return eval(compile(src, '<stdin>', 'eval'), *args, **kwargs)


class ReprEncoder(JSONEncoder):
    """JSON encoder using repr for objects"""

    def default(self, obj):
        return repr(obj)


def dump(o):
    """Shortcut to json.dumps with ReprEncoder"""
    return dumps(o, cls=ReprEncoder, sort_keys=True)


def tokenize_redir(raw_data):
    raw_io = StringIO()
    raw_io.write(raw_data)
    raw_io.seek(0)
    last_token = ''

    for token_type, token, src, erc, line in generate_tokens(raw_io.readline):
        if (
            token_type == tokens.ERRORTOKEN
            and token == '!'
            and last_token in ('>', '>>')
        ):
            return (
                line[: src[1] - 1],
                line[erc[1] :].lstrip(),
                last_token == '>>',
            )
        last_token = token
    return


class Interaction(object):

    hooks = {
        'update_watchers': [
            'start',
            'eval',
            'watch',
            'init',
            'select',
            'unwatch',
        ]
    }

    def __init__(
        self,
        db,
        frame,
        tb,
        exception,
        exception_description,
        init=None,
        shell=False,
        shell_vars=None,
        source=None,
        timeout=None,
    ):
        self.db = db
        self.shell = shell
        self.init_message = init
        self.stack, self.trace, self.index = self.db.get_trace(frame, tb)
        self.exception = exception
        self.exception_description = exception_description
        # Copy locals to avoid strange cpython behaviour
        self.locals = list(map(lambda x: x[0].f_locals, self.stack))
        self.htmldiff = Html5Diff(4)
        self.timeout = timeout

        if self.shell:
            self.locals[self.index] = shell_vars or {}

        if source:
            with open(source) as f:
                compiled_code = compile(f.read(), '<source>', 'exec')
            # Executing in locals to keep local scope
            # (http://bugs.python.org/issue16781)
            execute(compiled_code, self.current_locals, self.current_locals)

    def hook(self, kind):
        for hook, events in self.hooks.items():
            if kind in events:
                getattr(self, hook)()

    @property
    def current(self):
        return self.trace[self.index]

    @property
    def current_frame(self):
        return self.stack[self.index][0]

    @property
    def current_locals(self):
        return self.locals[self.index]

    @property
    def current_file(self):
        return self.current['file']

    def get_globals(self):
        """Get enriched globals"""
        if self.shell:
            globals_ = dict(_initial_globals)
        else:
            globals_ = dict(self.current_frame.f_globals)
        globals_['_'] = self.db.last_obj
        if cut is not None:
            globals_.setdefault('cut', cut)
        # For meta debuging purpose
        globals_['___wdb'] = self.db
        # Hack for function scope eval
        globals_.update(self.current_locals)
        for var, val in self.db.extra_vars.items():
            globals_[var] = val
        self.db.extra_items = {}
        return globals_

    def init(self):
        self.db.send(
            'Title|%s'
            % dump(
                {
                    'title': self.exception,
                    'subtitle': self.exception_description,
                }
            )
        )
        if self.shell:
            self.db.send('Shell')
        else:
            self.db.send(
                'Trace|%s' % dump({'trace': self.trace, 'cwd': os.getcwd()})
            )
            self.db.send(
                'SelectCheck|%s'
                % dump({'frame': self.current, 'name': self.current_file})
            )
        if self.init_message:
            self.db.send(self.init_message)
            self.init_message = None
        self.hook('init')

    def parse_command(self, message):
        # Parse received message
        if '|' in message:
            return message.split('|', 1)
        return message, ''

    def loop(self):
        stop = False
        while not stop:
            self.db.send('UPDATE_FILENAME|%s' % self.current_file)
            try:
                stop = self.interact()
            except Exception:
                log.exception('Error in loop')
                try:
                    exc = self.handle_exc()
                    type_, value = sys.exc_info()[:2]
                    link = (
                        '<a href="https://github.com/Kozea/wdb/issues/new?'
                        'title=%s&body=%s&labels=defect" class="nogood">'
                        'Please click here to report it on Github</a>'
                    ) % (
                        quote('%s: %s' % (type_.__name__, str(value))),
                        quote('```\n%s\n```\n' % traceback.format_exc()),
                    )
                    self.db.send(
                        'Echo|%s'
                        % dump(
                            {
                                'for': 'Error in Wdb, this is bad',
                                'val': exc + '<br>' + link,
                            }
                        )
                    )
                except Exception:
                    log.exception('Error in loop exception handling')
                    self.db.send(
                        'Echo|%s'
                        % dump(
                            {
                                'for': 'Too many errors',
                                'val': (
                                    "Don't really know what to say. "
                                    "Maybe it will work tomorrow."
                                ),
                            }
                        )
                    )

    def interact(self):
        try:
            message = self.db.receive(self.timeout)
            # Only timeout at first request
            self.timeout = None
        except KeyboardInterrupt:
            # Quit on KeyboardInterrupt
            message = 'Quit'

        cmd, data = self.parse_command(message)
        cmd = cmd.lower()
        log.debug('Cmd %s #Data %d' % (cmd, len(data)))
        fun = getattr(self, 'do_' + cmd, None)
        if fun:
            rv = fun(data)
            self.hook(cmd)
            return rv

        log.warning('Unknown command %s' % cmd)

    def update_watchers(self):
        watched = {}
        for watcher in self.db.watchers[self.current_file]:
            try:
                watched[watcher] = self.db.safe_better_repr(
                    eval_(watcher, self.get_globals(), self.current_locals)
                )
            except Exception as e:
                watched[watcher] = type(e).__name__

        self.db.send('Watched|%s' % dump(watched))

    def notify_exc(self, msg):
        log.info(msg, exc_info=True)
        self.db.send(
            'Log|%s'
            % dump({'message': '%s\n%s' % (msg, traceback.format_exc())})
        )

    def do_start(self, data):
        self.started = True
        # Getting breakpoints
        log.debug('Getting breakpoints')

        self.db.send(
            'Init|%s'
            % dump(
                {
                    'cwd': os.getcwd(),
                    'version': __version__,
                    'breaks': self.db.breakpoints_to_json(),
                }
            )
        )
        self.db.send(
            'Title|%s'
            % dump(
                {
                    'title': self.exception,
                    'subtitle': self.exception_description,
                }
            )
        )
        if self.shell:
            self.db.send('Shell')
        else:
            self.db.send('Trace|%s' % dump({'trace': self.trace}))

            # In case of exception always be at top frame to start
            self.index = len(self.stack) - 1
            self.db.send(
                'SelectCheck|%s'
                % dump({'frame': self.current, 'name': self.current_file})
            )

        if self.init_message:
            self.db.send(self.init_message)
            self.init_message = None

    def do_select(self, data):
        self.index = int(data)
        self.db.send(
            'SelectCheck|%s'
            % dump({'frame': self.current, 'name': self.current_file})
        )

    def do_file(self, data):
        fn = data
        file = self.db.get_file(fn)
        self.db.send(
            'Select|%s'
            % dump({'frame': self.current, 'name': fn, 'file': file})
        )

    def do_inspect(self, data):
        if '/' in data:
            mode, data = data.split('/', 1)
        else:
            mode = 'inspect'

        try:
            thing = self.db.obj_cache.get(int(data))
        except Exception:
            self.fail('Inspect')
            return
        if mode == 'dump':
            self.db.send(
                'Print|%s'
                % dump(
                    {
                        'for': self.db.safe_better_repr(thing, html=False),
                        'result': self.db.safe_better_repr(thing, full=True),
                    }
                )
            )
            return

        if isinstance(thing, tuple) and len(thing) == 3:
            type_, value, tb = thing
            iter_tb = tb
            while iter_tb.tb_next is not None:
                iter_tb = iter_tb.tb_next

            self.db.extra_vars['__recursive_exception__'] = value

            self.db.interaction(
                iter_tb.tb_frame, tb, type_.__name__, str(value)
            )
            return

        self.db.send(
            'Dump|%s'
            % dump(
                {
                    'for': self.db.safe_repr(thing),
                    'val': self.db.dmp(thing),
                    'doc': get_doc(thing),
                    'source': get_source(thing),
                }
            )
        )

    def do_dump(self, data):
        try:
            thing = eval_(data, self.get_globals(), self.current_locals)
        except Exception:
            self.fail('Dump')
            return

        self.db.send(
            'Dump|%s'
            % dump(
                {
                    'for': u('%s ⟶ %s ') % (data, self.db.safe_repr(thing)),
                    'val': self.db.dmp(thing),
                    'doc': get_doc(thing),
                    'source': get_source(thing),
                }
            )
        )

    def do_trace(self, data):
        self.db.send('Trace|%s' % dump({'trace': self.trace}))

    def do_eval(self, data):
        redir = None
        imports = []
        raw_data = data.strip()
        if raw_data.startswith('!<'):
            filename = raw_data[2:].strip()
            try:
                with open(filename, 'r') as f:
                    raw_data = f.read()
            except Exception:
                self.fail('Eval', 'Unable to read from file %s' % filename)
                return

        lines = raw_data.split('\n')
        if '>!' in lines[-1]:
            try:
                last_line, redir, append = tokenize_redir(raw_data)
            except TokenError:
                last_line = redir = None
                append = False
            if redir and last_line:
                indent = len(lines[-1]) - len(lines[-1].lstrip())
                lines[-1] = indent * u(' ') + last_line
                raw_data = '\n'.join(lines)
        data = raw_data
        # Keep spaces
        raw_data = raw_data.replace(' ', u(' '))
        # Compensate prompt for multi line
        raw_data = raw_data.replace('\n', '\n' + u(' ' * 4))
        duration = None
        with self.db.capture_output(with_hook=redir is None) as (out, err):
            compiled_code = None
            try:
                compiled_code = compile(data, '<stdin>', 'single')
            except Exception:
                try:
                    compiled_code = compile(data, '<stdin>', 'exec')
                except Exception:
                    maybe_hook = self.handle_exc()

                # Hack from codeop
                e1 = e2 = None
                try:
                    compiled_code = compile(data + '\n', '<stdin>', 'exec')
                except Exception as e:
                    e1 = e
                try:
                    compile(data + '\n\n', '<stdin>', 'exec')
                except Exception as e:
                    e2 = e

                if not compiled_code:
                    if repr(e1) != repr(e2):
                        # Multiline not terminated
                        self.db.send('NewLine')
                        return
                    else:
                        self.db.hooked = maybe_hook

            loc = self.current_locals
            start = time.time()
            if compiled_code is not None:
                self.db.compile_cache[id(compiled_code)] = data
                try:
                    execute(compiled_code, self.get_globals(), loc)
                except NameError as e:
                    m = re.match("name '(.+)' is not defined", str(e))
                    if m:
                        name = m.groups()[0]
                        if self.db._importmagic_index:
                            scores = self.db._importmagic_index.symbol_scores(
                                name
                            )
                            for _, module, variable in scores:
                                if variable is None:
                                    imports.append('import %s' % module)
                                else:
                                    imports.append(
                                        'from %s import %s'
                                        % (module, variable)
                                    )
                        elif importable_module(name):
                            imports.append('import %s' % name)

                    self.db.hooked = self.handle_exc()
                except Exception:
                    self.db.hooked = self.handle_exc()

        duration = int((time.time() - start) * 1000 * 1000)
        if redir and not self.db.hooked:
            try:
                with open(redir, 'a' if append else 'w') as f:
                    f.write('\n'.join(out) + '\n'.join(err) + '\n')
            except Exception:
                self.fail('Eval', 'Unable to write to file %s' % redir)
                return
            self.db.send(
                'Print|%s'
                % dump(
                    {
                        'for': raw_data,
                        'result': escape(
                            '%s to file %s'
                            % ('Appended' if append else 'Written', redir)
                        ),
                    }
                )
            )
        else:
            rv = escape('\n'.join(out) + '\n'.join(err))
            try:
                dump(rv)
            except Exception:
                rv = rv.decode('ascii', 'ignore')

            if rv and self.db.hooked:
                result = self.db.hooked + '\n' + rv
            elif rv:
                result = rv
            else:
                result = self.db.hooked

            self.db.send(
                'Print|%s'
                % dump(
                    {'for': raw_data, 'result': result, 'duration': duration}
                )
            )
            if imports:
                self.db.send('Suggest|%s' % dump({'imports': imports}))

    def do_ping(self, data):
        self.db.send('Pong')

    def do_step(self, data):
        self.db.set_step(self.current_frame)
        return True

    def do_next(self, data):
        self.db.set_next(self.current_frame)
        return True

    def do_continue(self, data):
        self.db.stepping = False
        self.db.set_continue(self.current_frame)
        return True

    def do_return(self, data):
        self.db.set_return(self.current_frame)
        return True

    def do_until(self, data):
        self.db.set_until(self.current_frame)
        return True

    def do_close(self, data):
        self.db.stepping = False
        if self.db.closed is not None:
            # Ignore set_trace till end
            self.db.closed = True

        self.db.set_continue(self.current_frame)
        return True

    def do_break(self, data):
        from linecache import getline

        brk = loads(data)

        def break_fail(x):
            return self.fail(
                'Break',
                'Break on %s failed' % ('%s:%s' % (brk['fn'], brk['lno'])),
                message=x,
            )

        if not brk.get('fn'):
            break_fail('Can’t break with no current file')
            return

        if brk['lno'] is not None:
            try:
                lno = int(brk['lno'])
            except Exception:
                break_fail(
                    'Wrong breakpoint format must be '
                    '[file][:lineno][#function][,condition].'
                )
                return

            line = getline(brk['fn'], lno, self.current_frame.f_globals)
            if not line:
                for path in sys.path:
                    line = getline(
                        os.path.join(path, brk['fn']),
                        brk['lno'],
                        self.current_frame.f_globals,
                    )
                    if line:
                        break
            if not line:
                break_fail('Line does not exist')
                return

            if not executable_line(line):
                break_fail('Blank line or comment')
                return

        breakpoint = self.db.set_break(
            brk['fn'], brk['lno'], brk['temporary'], brk['cond'], brk['fun']
        )
        break_set = breakpoint.to_dict()
        break_set['temporary'] = brk['temporary']
        self.db.send('BreakSet|%s' % dump(break_set))

    def do_unbreak(self, data):
        brk = loads(data)
        lno = brk['lno'] and int(brk['lno'])
        self.db.clear_break(
            brk['fn'], lno, brk['temporary'], brk['cond'], brk['fun']
        )

        self.db.send('BreakUnset|%s' % data)

    def do_breakpoints(self, data):
        self.db.send(
            'Print|%s'
            % dump({'for': 'Breakpoints', 'result': self.db.breakpoints})
        )

    def do_watch(self, data):
        self.db.watchers[self.current_file].add(data)
        self.db.send('Ack')

    def do_unwatch(self, data):
        self.db.watchers[self.current_file].remove(data)

    def do_jump(self, data):
        lno = int(data)
        if self.index != len(self.trace) - 1:
            log.error('Must be at bottom frame')
            return

        try:
            self.current_frame.f_lineno = lno
        except ValueError:
            self.fail('Unbreak')
            return

        self.current['lno'] = lno
        self.db.send('Trace|%s' % dump({'trace': self.trace}))
        self.db.send(
            'SelectCheck|%s'
            % dump({'frame': self.current, 'name': self.current_file})
        )

    def do_complete(self, data):
        completion = loads(data)
        manual = completion.pop('manual', False)
        if manual:
            timeout = 5
        else:
            timeout = 0.1

        source = completion.pop('source')
        pos = completion.pop('pos')
        if not Interpreter:
            self.db.send('Suggest')
            return
        try:
            script = Interpreter(
                source, [self.current_locals, self.get_globals()], **completion
            )
            with timeout_of(timeout, not manual):
                completions = script.completions()
        except Exception:
            self.db.send('Suggest')
            if log.level < WARNING:
                self.notify_exc('Completion failed for %s' % data)
            return

        try:
            with timeout_of(timeout / 2, not manual):
                funs = script.call_signatures() or []
        except Exception:
            self.db.send('Suggest')
            if log.level < WARNING:
                self.notify_exc('Completion of function failed for %s' % data)
            return

        before = source[:pos]
        after = source[pos:]
        like = ''
        if len(completions):
            completion = completions[0]
            base = completion.name[
                : len(completion.name) - len(completion.complete)
            ]
            if len(base):
                like = before[-len(base) :]
                if len(like):
                    before = before[: -len(like)]
        try:
            suggest_obj = {
                'data': {'start': before, 'end': after, 'like': like},
                'params': [
                    {
                        'params': [
                            p.description.replace('\n', '') for p in fun.params
                        ],
                        'index': fun.index,
                        'module': fun.module_name,
                        'call_name': fun.name,
                    }
                    for fun in funs
                ],
                'completions': [
                    {
                        'base': comp.name[
                            : len(comp.name) - len(comp.complete)
                        ],
                        'complete': comp.complete,
                        'description': comp.description,
                    }
                    for comp in completions
                    if comp.name.endswith(comp.complete)
                ],
            }
            self.db.send('Suggest|%s' % dump(suggest_obj))
        except Exception:
            self.db.send('Suggest')
            self.notify_exc('Completion generation failed for %s' % data)

    def do_save(self, data):
        fn, src = data.split('|', 1)
        if not os.path.exists(fn):
            return
        try:
            encoding = _detect_lines_encoding(src.splitlines())
            with inplace(fn, encoding=encoding) as (_, w):
                w.write(src)
        except Exception as e:
            self.db.send(
                'Echo|%s' % dump({'for': 'Error during save', 'val': str(e)})
            )
        else:
            self.db.send(
                'Echo|%s'
                % dump({'for': 'Save succesful', 'val': 'Wrote %s' % fn})
            )

    def do_external(self, data):
        default = {'linux': 'xdg-open', 'win32': '', 'darwin': 'open'}.get(
            sys.platform, 'open'
        )
        editor = os.getenv('EDITOR', os.getenv('VISUAL', default))
        if editor:
            cmd = editor.split(' ')
        else:
            cmd = []
        try:
            Popen(cmd + [data])
        except Exception:
            self.fail('External open')

    def do_display(self, data):
        if ';' in data:
            mime, data = data.split(';', 1)
            forced = True
        else:
            mime = 'text/html'
            forced = False

        try:
            thing = eval_(data, self.get_globals(), self.current_locals)
        except Exception:
            self.fail('Display')
            return
        else:
            thing = force_bytes(thing)
            if magic and not forced:
                mime = magic.from_buffer(thing, mime=True)
            self.db.send(
                'Display|%s'
                % dump(
                    {
                        'for': u('%s (%s)') % (data, mime),
                        'val': from_bytes(b64encode(thing)),
                        'type': mime,
                    }
                )
            )

    def do_disable(self, data):
        self.db.__class__.enabled = False
        self.db.stepping = False
        self.db.stop_trace()
        self.db.die()
        return True

    def do_quit(self, data):
        self.db.stepping = False
        self.db.stop_trace()
        return True

    def do_restart(self, data):
        try:
            # Try re-execing as-is
            os.execvp(sys.argv[0], sys.argv)
        except Exception:
            # Put the python executable in front
            python = sys.executable
            os.execl(python, python, *sys.argv)

    def do_diff(self, data):
        if '?' not in data and '<>' not in data:
            self.fail(
                'Diff',
                'Diff error',
                'You must provide two expression '
                'separated by "?" or "<>" to make a diff',
            )
            return
        pretty = '?' in data
        expressions = [
            expression.strip()
            for expression in (
                data.split('?') if '?' in data else data.split('<>')
            )
        ]
        strings = []
        for expression in expressions:
            try:
                strings.append(
                    eval_(expression, self.get_globals(), self.current_locals)
                )
            except Exception:
                self.fail(
                    'Diff',
                    "Diff failed: Expression %s "
                    "failed to evaluate to a string" % expression,
                )
                return

        render = (
            (
                (
                    lambda x: self.db.better_repr(x, html=False)
                    or self.db.safe_repr(x)
                )
            )
            if pretty
            else str
        )
        strings = [
            render(string) if not is_str(string) else string
            for string in strings
        ]
        self.db.send(
            'RawHTML|%s'
            % dump(
                {
                    'for': u('Difference between %s')
                    % (' and '.join(expressions)),
                    'val': self.htmldiff.make_table(
                        strings[0].splitlines(keepends=True),
                        strings[1].splitlines(keepends=True),
                        expressions[0],
                        expressions[1],
                    ),
                }
            )
        )

    def do_find(self, data):
        if ' in ' not in data and ' of ' not in data:
            self.fail(
                'Find',
                'Find error',
                'Syntax for find is: "key in expression" '
                'or "value testing function of expression"',
            )
        if ' in ' in data:
            key, expr = data.split(' in ')
        else:
            key, expr = data.split(' of ')

        try:
            value = eval_(expr, self.get_globals(), self.current_locals)
        except Exception:
            self.fail('Find')
            return
        if ' in ' in data:
            matches = search_key_in_obj(key, value, path='%s.' % expr)
        else:
            matches = search_value_in_obj(key, value, path='%s.' % expr)

        self.db.send(
            'Print|%s'
            % dump(
                {
                    'for': 'Finding %s in %s' % (key, expr),
                    'result': 'Found:\n%s'
                    % '\n'.join(
                        [
                            '%s: -> %s' % (k, escape(self.db.safe_repr(val)))
                            for k, val in matches
                        ]
                    )
                    if matches
                    else 'Not found',
                }
            )
        )

    def handle_exc(self):
        """Return a formated exception traceback for wdb.js use"""
        exc_info = sys.exc_info()
        type_, value = exc_info[:2]
        self.db.obj_cache[id(exc_info)] = exc_info

        return '<a href="%d" class="inspect">%s: %s</a>' % (
            id(exc_info),
            escape(type_.__name__),
            escape(repr(value)),
        )

    def fail(self, cmd, title=None, message=None):
        """Send back captured exceptions"""
        if message is None:
            message = self.handle_exc()
        else:
            message = escape(message)
        self.db.send(
            'Echo|%s'
            % dump({'for': escape(title or '%s failed' % cmd), 'val': message})
        )



# *-* coding: utf-8 *-*
# This file is part of wdb
#
# wdb Copyright (c) 2012-2016  Florian Mounier, Kozea
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import with_statement

try:
    import pkg_resources
except ImportError:
    __version__ = "pkg_resources not found on PYTHON_PATH"
else:
    try:
        __version__ = pkg_resources.require('wdb')[0].version
    except pkg_resources.DistributionNotFound:
        __version__ = "wdb is not installed"

_initial_globals = dict(globals())

from ._compat import (
    execute,
    StringIO,
    to_unicode_string,
    escape,
    loads,
    JSONDecodeError,
    Socket,
    logger,
    OrderedDict,
)

from .breakpoint import (
    Breakpoint,
    LineBreakpoint,
    ConditionalBreakpoint,
    FunctionBreakpoint,
)

from collections import defaultdict
from functools import wraps
from .ui import Interaction, dump
from .utils import (
    pretty_frame,
    executable_line,
    get_args,
    get_source_from_byte_code,
    cut_if_too_long,
    IterableEllipsis,
)
from .state import Running, Step, Next, Until, Return
from contextlib import contextmanager
from uuid import uuid4
from threading import Thread
import dis
import os
import logging
import sys
import threading
import socket
import webbrowser
import atexit
import time

try:
    import importmagic
except ImportError:
    importmagic = None

# Get wdb server host
SOCKET_SERVER = os.getenv('WDB_SOCKET_SERVER', 'localhost')
# and port
SOCKET_PORT = int(os.getenv('WDB_SOCKET_PORT', '19840'))

# Get wdb web server host
WEB_SERVER = os.getenv('WDB_WEB_SERVER')
# and port
WEB_PORT = int(os.getenv('WDB_WEB_PORT', 0))

WDB_NO_BROWSER_AUTO_OPEN = bool(os.getenv('WDB_NO_BROWSER_AUTO_OPEN', False))
log = logger('wdb')
trace_log = logging.getLogger('wdb.trace')

for log_name in ('main', 'trace', 'ui', 'ext', 'bp'):
    logger_name = 'wdb.%s' % log_name if log_name != 'main' else 'wdb'
    level = os.getenv(
        'WDB_%s_LOG' % log_name.upper(), os.getenv('WDB_LOG', 'WARNING')
    ).upper()
    logging.getLogger(logger_name).setLevel(getattr(logging, level, 'WARNING'))


class Wdb(object):
    """Wdb debugger main class"""

    _instances = {}
    _sockets = []
    enabled = True
    breakpoints = set()
    watchers = defaultdict(set)

    @staticmethod
    def get(no_create=False, server=None, port=None, force_uuid=None):
        """Get the thread local singleton"""
        pid = os.getpid()
        thread = threading.current_thread()
        wdb = Wdb._instances.get((pid, thread))
        if not wdb and not no_create:
            wdb = object.__new__(Wdb)
            Wdb.__init__(wdb, server, port, force_uuid)
            wdb.pid = pid
            wdb.thread = thread
            Wdb._instances[(pid, thread)] = wdb
        elif wdb:
            if (
                server is not None
                and wdb.server != server
                or port is not None
                and wdb.port != port
            ):
                log.warn('Different server/port set, ignoring')
            else:
                wdb.reconnect_if_needed()
        return wdb

    @staticmethod
    def pop():
        """Remove instance from instance list"""
        pid = os.getpid()
        thread = threading.current_thread()
        Wdb._instances.pop((pid, thread))

    def __new__(cls, server=None, port=None):
        return cls.get(server=server, port=port)

    def __init__(self, server=None, port=None, force_uuid=None):
        log.debug('New wdb instance %r' % self)
        self.obj_cache = {}
        self.compile_cache = {}
        self.tracing = False
        self.begun = False
        self.connected = False
        self.closed = None  # Handle request long ignores for ext
        self.stepping = False
        self.extra_vars = {}
        self.last_obj = None
        self.reset()
        self.uuid = force_uuid or str(uuid4())
        self.state = Running(None)
        self.full = False
        self.below = 0
        self.under = None
        self.server = server or SOCKET_SERVER
        self.port = port or SOCKET_PORT
        self.interaction_stack = []
        self._importmagic_index = None
        self._importmagic_index_lock = threading.RLock()
        self.index_imports()
        self._socket = None
        self.connect()
        self.get_breakpoints()

    def run_file(self, filename):
        """Run the file `filename` with trace"""
        import __main__

        __main__.__dict__.clear()
        __main__.__dict__.update(
            {
                "__name__": "__main__",
                "__file__": filename,
                "__builtins__": __builtins__,
            }
        )
        with open(filename, "rb") as fp:
            statement = compile(fp.read(), filename, 'exec')
        self.run(statement, filename)

    def run(self, cmd, fn=None, globals=None, locals=None):
        """Run the cmd `cmd` with trace"""
        if globals is None:
            import __main__

            globals = __main__.__dict__
        if locals is None:
            locals = globals
        self.reset()
        if isinstance(cmd, str):
            str_cmd = cmd
            cmd = compile(str_cmd, fn or "<wdb>", "exec")
            self.compile_cache[id(cmd)] = str_cmd
        if fn:
            from linecache import getline

            lno = 1
            while True:
                line = getline(fn, lno, globals)
                if line is None:
                    lno = None
                    break
                if executable_line(line):
                    break
                lno += 1

        self.start_trace()
        if lno is not None:
            self.breakpoints.add(LineBreakpoint(fn, lno, temporary=True))
        try:
            execute(cmd, globals, locals)
        finally:
            self.stop_trace()

    def reset(self):
        """Refresh linecache"""
        import linecache

        linecache.checkcache()

    def reconnect_if_needed(self):
        try:
            # Sending PING twice
            self.send('PING')
            self.send('PING')
            log.debug('Dual ping sent')
        except socket.error:
            log.warning('socket error on ping, connection lost retrying')
            self._socket = None
            self.connected = False
            self.begun = False
            self.connect()

    def connect(self):
        """Connect to wdb server"""
        log.info('Connecting socket on %s:%d' % (self.server, self.port))
        tries = 0
        while not self._socket and tries < 10:
            try:
                time.sleep(0.2 * tries)
                self._socket = Socket((self.server, self.port))
            except socket.error:
                tries += 1
                log.warning(
                    'You must start/install wdb.server '
                    '(Retrying on %s:%d) [Try #%d/10]'
                    % (self.server, self.port, tries)
                )
                self._socket = None

        if not self._socket:
            log.warning('Could not connect to server')
            return

        Wdb._sockets.append(self._socket)
        self._socket.send_bytes(self.uuid.encode('utf-8'))

    def get_breakpoints(self):
        log.info('Getting server breakpoints')
        self.send('ServerBreaks')
        breaks = self.receive()
        try:
            breaks = loads(breaks)
        except JSONDecodeError:
            breaks = []
        self._init_breakpoints = breaks

        for brk in breaks:
            self.set_break(
                brk['fn'], brk['lno'], False, brk['cond'], brk['fun']
            )

        log.info('Server breakpoints added')

    def index_imports(self):
        if not importmagic or self._importmagic_index:
            return

        self._importmagic_index_lock.acquire()

        def index(self):
            log.info('Indexing imports')
            index = importmagic.SymbolIndex()
            index.build_index(sys.path)
            self._importmagic_index = index
            log.info('Indexing imports done')

        index_thread = Thread(
            target=index, args=(self,), name='wdb_importmagic_build_index'
        )
        # Don't wait for completion, let it die alone:
        index_thread.daemon = True
        index_thread.start()

        self._importmagic_index_lock.release()

    def breakpoints_to_json(self):
        return [brk.to_dict() for brk in self.breakpoints]

    def _get_under_code_ref(self):
        code = getattr(self.under, '__code__', None)
        if not code and hasattr(self.under, '__call__'):
            # Allow for callable objects
            code = getattr(self.under.__call__, '__code__', None)

        return code

    def _walk_frame_ancestry(self, frame):
        iframe = frame
        while iframe is not None:
            yield iframe
            iframe = iframe.f_back

    def check_below(self, frame):
        stop_frame = self.state.frame

        if not any((self.below, self.under)):
            return frame == stop_frame, False

        under_code = self._get_under_code_ref()
        if under_code:
            stop_frame = None
            for iframe in self._walk_frame_ancestry(frame):
                if iframe.f_code == under_code:
                    stop_frame = iframe

        if not stop_frame:
            return False, False

        below = 0
        for iframe in self._walk_frame_ancestry(frame):
            if stop_frame == iframe:
                break
            below += 1

        return below == self.below, below == self.below

    def trace_dispatch(self, frame, event, arg):
        """This function is called every line,
        function call, function return and exception during trace"""
        fun = getattr(self, 'handle_' + event, None)
        if not fun:
            return self.trace_dispatch
        below, continue_below = self.check_below(frame)
        if (
            self.state.stops(frame, event)
            or (event == 'line' and self.breaks(frame))
            or (event == 'exception' and (self.full or below))
        ):
            fun(frame, arg)

        if event == 'return' and frame == self.state.frame:
            # Upping state
            if self.state.up():
                # No more frames
                self.stop_trace()
                return
            # Threading / Multiprocessing support
            co = self.state.frame.f_code
            if (
                co.co_filename.endswith('threading.py')
                and co.co_name.endswith('_bootstrap_inner')
            ) or (
                self.state.frame.f_code.co_filename.endswith(
                    os.path.join('multiprocessing', 'process.py')
                )
                and self.state.frame.f_code.co_name == '_bootstrap'
            ):
                # Thread / Process is dead
                self.stop_trace()
                self.die()
                return

        if (
            event == 'call'
            and not self.stepping
            and not self.full
            and not continue_below
            and not self.get_file_breaks(frame.f_code.co_filename)
        ):
            # Don't trace anymore here
            return
        return self.trace_dispatch

    def trace_debug_dispatch(self, frame, event, arg):
        """Utility function to add debug to tracing"""
        trace_log.info(
            'Frame:%s. Event: %s. Arg: %r' % (pretty_frame(frame), event, arg)
        )
        trace_log.debug(
            'state %r breaks ? %s stops ? %s'
            % (
                self.state,
                self.breaks(frame, no_remove=True),
                self.state.stops(frame, event),
            )
        )
        if event == 'return':
            trace_log.debug(
                'Return: frame: %s, state: %s, state.f_back: %s'
                % (
                    pretty_frame(frame),
                    pretty_frame(self.state.frame),
                    pretty_frame(self.state.frame.f_back),
                )
            )
        if self.trace_dispatch(frame, event, arg):
            return self.trace_debug_dispatch
        trace_log.debug("No trace %s" % pretty_frame(frame))

    def start_trace(self, full=False, frame=None, below=0, under=None):
        """Start tracing from here"""
        if self.tracing:
            return
        self.reset()
        log.info('Starting trace')
        frame = frame or sys._getframe().f_back
        # Setting trace without pausing
        self.set_trace(frame, break_=False)
        self.tracing = True
        self.below = below
        self.under = under
        self.full = full

    def set_trace(self, frame=None, break_=True):
        """Break at current state"""
        # We are already tracing, do nothing
        trace_log.info(
            'Setting trace %s (stepping %s) (current_trace: %s)'
            % (
                pretty_frame(frame or sys._getframe().f_back),
                self.stepping,
                sys.gettrace(),
            )
        )
        if self.stepping or self.closed:
            return
        self.reset()
        trace = (
            self.trace_dispatch
            if trace_log.level >= 30
            else self.trace_debug_dispatch
        )
        trace_frame = frame = frame or sys._getframe().f_back
        while frame:
            frame.f_trace = trace
            frame = frame.f_back
        self.state = Step(trace_frame) if break_ else Running(trace_frame)
        sys.settrace(trace)

    def stop_trace(self, frame=None):
        """Stop tracing from here"""
        self.tracing = False
        self.full = False
        frame = frame or sys._getframe().f_back
        while frame:
            del frame.f_trace
            frame = frame.f_back
        sys.settrace(None)
        log.info('Stopping trace')

    def set_until(self, frame, lineno=None):
        """Stop on the next line number."""
        self.state = Until(frame, frame.f_lineno)

    def set_step(self, frame):
        """Stop on the next line."""
        self.state = Step(frame)

    def set_next(self, frame):
        """Stop on the next line in current frame."""
        self.state = Next(frame)

    def set_return(self, frame):
        """Stop when returning from the given frame."""
        self.state = Return(frame)

    def set_continue(self, frame):
        """Don't stop anymore"""
        self.state = Running(frame)
        if not self.tracing and not self.breakpoints:
            # If we were in a set_trace and there's no breakpoint to trace for
            # Run without trace
            self.stop_trace()

    def get_break(self, filename, lineno, temporary, cond, funcname):
        if lineno and not cond:
            return LineBreakpoint(filename, lineno, temporary)
        elif cond:
            return ConditionalBreakpoint(filename, lineno, cond, temporary)
        elif funcname:
            return FunctionBreakpoint(filename, funcname, temporary)
        else:
            return Breakpoint(filename, temporary)

    def set_break(
        self, filename, lineno=None, temporary=False, cond=None, funcname=None
    ):
        """Put a breakpoint for filename"""
        log.info(
            'Setting break fn:%s lno:%s tmp:%s cond:%s fun:%s'
            % (filename, lineno, temporary, cond, funcname)
        )
        breakpoint = self.get_break(
            filename, lineno, temporary, cond, funcname
        )
        self.breakpoints.add(breakpoint)
        log.info('Breakpoint %r added' % breakpoint)
        return breakpoint

    def clear_break(
        self, filename, lineno=None, temporary=False, cond=None, funcname=None
    ):
        """Remove a breakpoint"""
        log.info(
            'Removing break fn:%s lno:%s tmp:%s cond:%s fun:%s'
            % (filename, lineno, temporary, cond, funcname)
        )

        breakpoint = self.get_break(
            filename, lineno, temporary or False, cond, funcname
        )
        if temporary is None and breakpoint not in self.breakpoints:
            breakpoint = self.get_break(filename, lineno, True, cond, funcname)

        try:
            self.breakpoints.remove(breakpoint)
            log.info('Breakpoint %r removed' % breakpoint)
        except Exception:
            log.info('Breakpoint %r not removed: not found' % breakpoint)

    def safe_repr(self, obj):
        """Like a repr but without exception"""
        try:
            return repr(obj)
        except Exception as e:
            return '??? Broken repr (%s: %s)' % (type(e).__name__, e)

    def safe_better_repr(
        self, obj, context=None, html=True, level=0, full=False
    ):
        """Repr with inspect links on objects"""
        context = context and dict(context) or {}
        recursion = id(obj) in context
        if not recursion:
            context[id(obj)] = obj
            try:
                rv = self.better_repr(obj, context, html, level + 1, full)
            except Exception:
                rv = None
            if rv:
                return rv

        self.obj_cache[id(obj)] = obj
        if html:
            return '<a href="%d" class="inspect">%s%s</a>' % (
                id(obj),
                'Recursion of ' if recursion else '',
                escape(self.safe_repr(obj)),
            )
        return '%s%s' % (
            'Recursion of ' if recursion else '',
            self.safe_repr(obj),
        )

    def better_repr(self, obj, context=None, html=True, level=1, full=False):
        """Repr with html decorations or indentation"""
        abbreviate = (lambda x, level, **kw: x) if full else cut_if_too_long

        def get_too_long_repr(ie):
            r = '[%d more…]' % ie.size
            if html:
                self.obj_cache[id(obj)] = obj
                return '<a href="dump/%d" class="inspect">%s</a>' % (
                    id(obj),
                    r,
                )
            return r

        if isinstance(obj, dict):
            if isinstance(obj, OrderedDict):
                dict_sorted = lambda it, f: it
            else:
                dict_sorted = sorted

            dict_repr = '  ' * (level - 1)
            if type(obj) != dict:
                dict_repr = type(obj).__name__ + '({'
                closer = '})'
            else:
                dict_repr = '{'
                closer = '}'
            if len(obj) > 2:
                dict_repr += '\n' + '  ' * level
                if html:
                    dict_repr += '''<table class="
                        mdl-data-table mdl-js-data-table
                        mdl-data-table--selectable mdl-shadow--2dp">'''
                    dict_repr += ''.join(
                        [
                            (
                                '<tr><td class="key">'
                                + self.safe_repr(key)
                                + ':'
                                + '</td>'
                                '<td class="val '
                                + 'mdl-data-table__cell--non-numeric">'
                                + self.safe_better_repr(
                                    val, context, html, level, full
                                )
                                + '</td></tr>'
                            )
                            if not isinstance(key, IterableEllipsis)
                            else (
                                '<tr><td colspan="2" class="ellipse">'
                                + get_too_long_repr(key)
                                + '</td></tr>'
                            )
                            for key, val in abbreviate(
                                dict_sorted(obj.items(), key=lambda x: x[0]),
                                level,
                                tuple_=True,
                            )
                        ]
                    )
                    dict_repr += '</table>'
                else:
                    dict_repr += ('\n' + '  ' * level).join(
                        [
                            self.safe_repr(key)
                            + ': '
                            + self.safe_better_repr(
                                val, context, html, level, full
                            )
                            if not isinstance(key, IterableEllipsis)
                            else get_too_long_repr(key)
                            for key, val in abbreviate(
                                dict_sorted(obj.items(), key=lambda x: x[0]),
                                level,
                                tuple_=True,
                            )
                        ]
                    )
                closer = '\n' + '  ' * (level - 1) + closer
            else:
                dict_repr += ', '.join(
                    [
                        self.safe_repr(key)
                        + ': '
                        + self.safe_better_repr(
                            val, context, html, level, full
                        )
                        for key, val in dict_sorted(
                            obj.items(), key=lambda x: x[0]
                        )
                    ]
                )
            dict_repr += closer
            return dict_repr

        if any(
            [
                isinstance(obj, list),
                isinstance(obj, set),
                isinstance(obj, tuple),
            ]
        ):
            iter_repr = '  ' * (level - 1)
            if type(obj) == list:
                iter_repr = '['
                closer = ']'
            elif type(obj) == set:
                iter_repr = '{'
                closer = '}'
            elif type(obj) == tuple:
                iter_repr = '('
                closer = ')'
            else:
                iter_repr = escape(obj.__class__.__name__) + '(['
                closer = '])'

            splitter = ', '
            if len(obj) > 2 and html:
                splitter += '\n' + '  ' * level
                iter_repr += '\n' + '  ' * level
                closer = '\n' + '  ' * (level - 1) + closer

            iter_repr += splitter.join(
                [
                    self.safe_better_repr(val, context, html, level, full)
                    if not isinstance(val, IterableEllipsis)
                    else get_too_long_repr(val)
                    for val in abbreviate(obj, level)
                ]
            )

            iter_repr += closer
            return iter_repr

    @contextmanager
    def capture_output(self, with_hook=True):
        """Steal stream output, return them in string, restore them"""
        self.hooked = ''

        def display_hook(obj):
            # That's some dirty hack
            self.hooked += self.safe_better_repr(obj)
            self.last_obj = obj

        stdout, stderr = sys.stdout, sys.stderr
        if with_hook:
            d_hook = sys.displayhook
            sys.displayhook = display_hook

        sys.stdout, sys.stderr = StringIO(), StringIO()
        out, err = [], []
        try:
            yield out, err
        finally:
            out.extend(sys.stdout.getvalue().splitlines())
            err.extend(sys.stderr.getvalue().splitlines())
            if with_hook:
                sys.displayhook = d_hook

            sys.stdout, sys.stderr = stdout, stderr

    def dmp(self, thing):
        """Dump the content of an object in a dict for wdb.js"""

        def safe_getattr(key):
            """Avoid crash on getattr"""
            try:
                return getattr(thing, key)
            except Exception as e:
                return 'Error getting attr "%s" from "%s" (%s: %s)' % (
                    key,
                    thing,
                    type(e).__name__,
                    e,
                )

        return dict(
            (
                escape(key),
                {
                    'val': self.safe_better_repr(safe_getattr(key)),
                    'type': type(safe_getattr(key)).__name__,
                },
            )
            for key in dir(thing)
        )

    def get_file(self, filename):
        """Get file source from cache"""
        import linecache

        # Hack for frozen importlib bootstrap
        if filename == '<frozen importlib._bootstrap>':
            filename = os.path.join(
                os.path.dirname(linecache.__file__),
                'importlib',
                '_bootstrap.py',
            )
        return to_unicode_string(
            ''.join(linecache.getlines(filename)), filename
        )

    def get_stack(self, f, t):
        """Build the stack from frame and traceback"""
        stack = []
        if t and t.tb_frame == f:
            t = t.tb_next
        while f is not None:
            stack.append((f, f.f_lineno))
            f = f.f_back
        stack.reverse()
        i = max(0, len(stack) - 1)
        while t is not None:
            stack.append((t.tb_frame, t.tb_lineno))
            t = t.tb_next
        if f is None:
            i = max(0, len(stack) - 1)
        return stack, i

    def get_trace(self, frame, tb):
        """Get a dict of the traceback for wdb.js use"""
        import linecache

        frames = []
        stack, _ = self.get_stack(frame, tb)
        current = 0

        for i, (stack_frame, lno) in enumerate(stack):
            code = stack_frame.f_code
            filename = code.co_filename or '<unspecified>'
            line = None
            if filename[0] == '<' and filename[-1] == '>':
                line = get_source_from_byte_code(code)
                fn = filename
            else:
                fn = os.path.abspath(filename)
            if not line:
                linecache.checkcache(filename)
                line = linecache.getline(filename, lno, stack_frame.f_globals)
                if not line:
                    line = self.compile_cache.get(id(code), '')
                line = to_unicode_string(line, filename)
                line = line and line.strip()
            startlnos = dis.findlinestarts(code)
            lastlineno = list(startlnos)[-1][1]
            if frame == stack_frame:
                current = i
            frames.append(
                {
                    'file': fn,
                    'function': code.co_name,
                    'flno': code.co_firstlineno,
                    'llno': lastlineno,
                    'lno': lno,
                    'code': line,
                    'level': i,
                    'current': frame == stack_frame,
                }
            )

        # While in exception always put the context to the top
        return stack, frames, current

    def send(self, data):
        """Send data through websocket"""
        log.debug('Sending %s' % data)
        if not self._socket:
            log.warn('No connection')
            return
        self._socket.send_bytes(data.encode('utf-8'))

    def receive(self, timeout=None):
        """Receive data through websocket"""
        log.debug('Receiving')
        if not self._socket:
            log.warn('No connection')
            return
        try:
            if timeout:
                rv = self._socket.poll(timeout)
                if not rv:
                    log.info('Connection timeouted')
                    return 'Quit'

            data = self._socket.recv_bytes()
        except Exception:
            log.error('Connection lost')
            return 'Quit'
        log.debug('Got %s' % data)
        return data.decode('utf-8')

    def open_browser(self, type_='debug'):
        if not self.connected:
            log.debug('Launching browser and wait for connection')
            web_url = 'http://%s:%d/%s/session/%s' % (
                WEB_SERVER or 'localhost',
                WEB_PORT or 1984,
                type_,
                self.uuid,
            )

            server = WEB_SERVER or '[wdb.server]'
            if WEB_PORT:
                server += ':%s' % WEB_PORT

            if WDB_NO_BROWSER_AUTO_OPEN:
                log.warning(
                    'You can now launch your browser at '
                    'http://%s/%s/session/%s' % (server, type_, self.uuid)
                )

            elif not webbrowser.open(web_url):
                log.warning(
                    'Unable to open browser, '
                    'please go to http://%s/%s/session/%s'
                    % (server, type_, self.uuid)
                )

            self.connected = True

    def shell(self, source=None, vars=None):
        self.interaction(
            sys._getframe(),
            exception_description='Shell',
            shell=True,
            shell_vars=vars,
            source=source,
        )

    def interaction(
        self,
        frame,
        tb=None,
        exception='Wdb',
        exception_description='Stepping',
        init=None,
        shell=False,
        shell_vars=None,
        source=None,
        iframe_mode=False,
        timeout=None,
        post_mortem=False,
    ):
        """User interaction handling blocking on socket receive"""
        log.info(
            'Interaction %r %r %r %r'
            % (frame, tb, exception, exception_description)
        )
        self.reconnect_if_needed()
        self.stepping = not shell

        if not iframe_mode:
            opts = {}
            if shell:
                opts['type_'] = 'shell'
            if post_mortem:
                opts['type_'] = 'pm'
            self.open_browser(**opts)

        lvl = len(self.interaction_stack)
        if lvl:
            exception_description += ' [recursive%s]' % (
                '^%d' % lvl if lvl > 1 else ''
            )

        interaction = Interaction(
            self,
            frame,
            tb,
            exception,
            exception_description,
            init=init,
            shell=shell,
            shell_vars=shell_vars,
            source=source,
            timeout=timeout,
        )

        self.interaction_stack.append(interaction)

        # For meta debugging purpose
        self._ui = interaction

        if self.begun:
            # Each new state sends the trace and selects a frame
            interaction.init()
        else:
            self.begun = True

        interaction.loop()
        self.interaction_stack.pop()
        if lvl:
            self.interaction_stack[-1].init()

    def handle_call(self, frame, argument_list):
        """This method is called when there is the remote possibility
        that we ever need to stop in this function."""
        fun = frame.f_code.co_name
        log.info('Calling: %r' % fun)

        init = 'Echo|%s' % dump(
            {
                'for': '__call__',
                'val': '%s(%s)'
                % (
                    fun,
                    ', '.join(
                        [
                            '%s=%s' % (key, self.safe_better_repr(value))
                            for key, value in get_args(frame).items()
                        ]
                    ),
                ),
            }
        )
        self.interaction(
            frame, init=init, exception_description='Calling %s' % fun
        )

    def handle_line(self, frame, arg):
        """This function is called when we stop or break at this line."""
        log.info('Stopping at line %s' % pretty_frame(frame))
        self.interaction(frame)

    def handle_return(self, frame, return_value):
        """This function is called when a return trap is set here."""
        self.obj_cache[id(return_value)] = return_value
        self.extra_vars['__return__'] = return_value
        fun = frame.f_code.co_name
        log.info('Returning from %r with value: %r' % (fun, return_value))

        init = 'Echo|%s' % dump(
            {'for': '__return__', 'val': self.safe_better_repr(return_value)}
        )
        self.interaction(
            frame,
            init=init,
            exception_description='Returning from %s with value %s'
            % (fun, return_value),
        )

    def handle_exception(self, frame, exc_info):
        """This function is called if an exception occurs,
        but only if we are to stop at or just below this level."""
        type_, value, tb = exc_info
        # Python 3 is broken see http://bugs.python.org/issue17413
        _value = value
        if not isinstance(_value, BaseException):
            _value = type_(value)
        fake_exc_info = type_, _value, tb
        log.error('Exception during trace', exc_info=fake_exc_info)
        self.obj_cache[id(exc_info)] = exc_info
        self.extra_vars['__exception__'] = exc_info
        exception = type_.__name__
        exception_description = str(value)
        init = 'Echo|%s' % dump(
            {
                'for': '__exception__',
                'val': escape('%s: %s') % (exception, exception_description),
            }
        )
        # User exception is 4 frames away from exception
        frame = frame or sys._getframe().f_back.f_back.f_back.f_back
        self.interaction(
            frame, tb, exception, exception_description, init=init
        )

    def breaks(self, frame, no_remove=False):
        """Return True if there's a breakpoint at frame"""
        for breakpoint in set(self.breakpoints):
            if breakpoint.breaks(frame):
                if breakpoint.temporary and not no_remove:
                    self.breakpoints.remove(breakpoint)
                return True
        return False

    def get_file_breaks(self, filename):
        """List all file `filename` breakpoints"""
        return [
            breakpoint
            for breakpoint in self.breakpoints
            if breakpoint.on_file(filename)
        ]

    def get_breaks_lno(self, filename):
        """List all line numbers that have a breakpoint"""
        return list(
            filter(
                lambda x: x is not None,
                [
                    getattr(breakpoint, 'line', None)
                    for breakpoint in self.breakpoints
                    if breakpoint.on_file(filename)
                ],
            )
        )

    def die(self):
        """Time to quit"""
        log.info('Time to die')
        if self.connected:
            try:
                self.send('Die')
            except Exception:
                pass
        if self._socket:
            self._socket.close()
        self.pop()


def set_trace(frame=None, skip=0, server=None, port=None):
    """Set trace on current line, or on given frame"""
    frame = frame or sys._getframe().f_back
    for i in range(skip):
        if not frame.f_back:
            break
        frame = frame.f_back
    wdb = Wdb.get(server=server, port=port)
    wdb.set_trace(frame)
    return wdb


def start_trace(
    full=False, frame=None, below=0, under=None, server=None, port=None
):
    """Start tracing program at callee level
       breaking on exception/breakpoints"""
    wdb = Wdb.get(server=server, port=port)
    if not wdb.stepping:
        wdb.start_trace(full, frame or sys._getframe().f_back, below, under)
    return wdb


def stop_trace(frame=None, close_on_exit=False):
    """Stop tracing"""
    log.info('Stopping trace')
    wdb = Wdb.get(True)  # Do not create an istance if there's None
    if wdb and (not wdb.stepping or close_on_exit):
        log.info('Stopping trace')
        wdb.stop_trace(frame or sys._getframe().f_back)
        if close_on_exit:
            wdb.die()
    return wdb


class trace(object):
    def __init__(self, **kwargs):
        """Make a tracing context with `with trace():`"""
        self.kwargs = kwargs

    def __enter__(self):
        #  2 calls to get here
        kwargs = dict(self.kwargs)
        if 'close_on_exit' in kwargs:
            kwargs.pop('close_on_exit')
        kwargs.setdefault('frame', sys._getframe().f_back)
        start_trace(**kwargs)

    def __exit__(self, *args):
        kwargs = {}
        kwargs['frame'] = self.kwargs.get('frame', sys._getframe().f_back)
        kwargs['close_on_exit'] = self.kwargs.get('close_on_exit', False)
        stop_trace(**kwargs)


def with_trace(fun):
    @wraps(fun)
    def traced(*args, **kwargs):
        with trace():
            return fun(*args, **kwargs)

    return traced


@atexit.register
def cleanup():
    """Close all sockets at exit"""
    for sck in list(Wdb._sockets):
        try:
            sck.close()
        except Exception:
            log.warn('Error in cleanup', exc_info=True)


def shell(source=None, vars=None, server=None, port=None):
    """Start a shell sourcing source or using vars as locals"""
    Wdb.get(server=server, port=port).shell(source=source, vars=vars)


# Pdb compatibility


def post_mortem(t=None, server=None, port=None):
    if t is None:
        t = sys.exc_info()[2]
        if t is None:
            raise ValueError(
                "A valid traceback must be passed if no "
                "exception is being handled"
            )

    wdb = Wdb.get(server=server, port=port)
    wdb.reset()
    wdb.interaction(None, t, post_mortem=True)


def pm(server=None, port=None):
    post_mortem(sys.last_traceback, server=server, port=port)



from .utils import pretty_frame


class State(object):
    def __init__(self, frame):
        self.frame = frame

    def up(self):
        """Go up in stack and return True if top frame"""
        if self.frame:
            self.frame = self.frame.f_back
            return self.frame is None

    def __repr__(self):
        return '<State is %s for %s>' % (
            self.__class__.__name__,
            pretty_frame(self.frame),
        )


class Running(State):
    """Running state: never stopping"""

    def stops(self, frame, event):
        return False


class Step(State):
    """Stepping state: always stopping"""

    def stops(self, frame, event):
        return True


class Next(State):
    """Nexting state: stop if same frame"""

    def stops(self, frame, event):
        return self.frame == frame


class Until(State):
    """Nexting until state: stop if same frame and is next line"""

    def __init__(self, frame, lno):
        self.frame = frame
        self.lno = lno + 1

    def stops(self, frame, event):
        return self.frame == frame and frame.f_lineno >= self.lno


class Return(Next):
    """Returning state: Stop on return event if same frame"""

    def stops(self, frame, event):
        return self.frame == frame and event == 'return'



import codecs
import re
import sys

python_version = sys.version_info[0]

try:
    from json import loads, dumps, JSONEncoder, JSONDecodeError
except ImportError:
    from simplejson import loads, dumps, JSONEncoder, JSONDecodeError

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote

try:
    from html import escape
except ImportError:
    from cgi import escape

try:
    from socketserver import TCPServer
except ImportError:
    from SocketServer import TCPServer

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

if python_version == 2:
    from StringIO import StringIO
else:
    from io import StringIO

if python_version == 2:

    def execute(cmd, globals_, locals_):
        exec('exec cmd in globals_, locals_')


else:

    def execute(cmd, globals_, locals_):
        exec(cmd, globals_, locals_)


_cookie_search = re.compile(r"coding[:=]\s*([-\w.]+)").search


def _detect_encoding(filename):
    import linecache

    lines = linecache.getlines(filename)
    return _detect_lines_encoding(lines)


def _detect_lines_encoding(lines):
    if not lines or lines[0].startswith(u("\xef\xbb\xbf")):
        return "utf-8"
    magic = _cookie_search("".join(lines[:2]))
    if magic is None:
        return 'utf-8'
    encoding = magic.group(1)
    try:
        codecs.lookup(encoding)
    except LookupError:
        return 'utf-8'
    return encoding


if python_version == 2:
    basestr = basestring

    def to_unicode(string):
        return string.decode('utf-8')

    def to_unicode_string(string, filename):
        if isinstance(string, unicode):
            return string

        encoding = _detect_encoding(filename)
        if encoding != 'utf-8' and string:
            return string.decode(encoding).encode('utf-8')
        else:
            return string

    def to_bytes(string):
        return string

    def from_bytes(bytes_):
        return bytes_

    def force_bytes(bytes_):
        if isinstance(bytes_, unicode):
            return bytes_.encode('utf-8')
        return bytes_


else:
    basestr = (str, bytes)

    def to_unicode(string):
        return string

    def to_unicode_string(string, filename):
        return string

    def to_bytes(string):
        return string.encode('utf-8')

    def from_bytes(bytes_):
        return bytes_.decode('utf-8')

    def force_bytes(bytes_):
        if isinstance(bytes_, str):
            return bytes_.encode('utf-8')
        return bytes_


def is_str(string):
    return isinstance(string, basestr)


def u(s):
    if python_version == 2:
        return s.decode('utf-8')
    return s


if python_version == 2:
    import struct
    import socket
    import errno
    import time
    import select

    has_winapi = False
    try:
        import _winapi

        has_winapi = True
    except ImportError:
        pass

    if sys.platform == 'win32' and has_winapi:
        from _winapi import WAIT_OBJECT_0, WAIT_TIMEOUT, INFINITE

        try:
            from _winapi import WAIT_ABANDONED_0
        except ImportError:
            WAIT_ABANDONED_0 = 128

        def _exhaustive_wait(handles, timeout):
            # Return ALL handles which are currently signalled.  (Only
            # returning the first signalled might create starvation issues.)
            L = list(handles)
            ready = []
            while L:
                res = _winapi.WaitForMultipleObjects(L, False, timeout)
                if res == WAIT_TIMEOUT:
                    break
                elif WAIT_OBJECT_0 <= res < WAIT_OBJECT_0 + len(L):
                    res -= WAIT_OBJECT_0
                elif WAIT_ABANDONED_0 <= res < WAIT_ABANDONED_0 + len(L):
                    res -= WAIT_ABANDONED_0
                else:
                    raise RuntimeError('Should not get here')
                ready.append(L[res])
                L = L[res + 1 :]
                timeout = 0
            return ready

        _ready_errors = set(
            (_winapi.ERROR_BROKEN_PIPE, _winapi.ERROR_NETNAME_DELETED)
        )

        def wait(object_list, timeout=None):
            '''
            Wait till an object in object_list is ready/readable.
            Returns list of those objects in object_list which are
            ready/readable.
            '''
            if timeout is None:
                timeout = INFINITE
            elif timeout < 0:
                timeout = 0
            else:
                timeout = int(timeout * 1000 + 0.5)

            object_list = list(object_list)
            waithandle_to_obj = {}
            ov_list = []
            ready_objects = set()
            ready_handles = set()

            try:
                for o in object_list:
                    try:
                        fileno = getattr(o, 'fileno')
                    except AttributeError:
                        waithandle_to_obj[o.__index__()] = o
                    else:
                        # start an overlapped read of length zero
                        try:
                            ov, err = _winapi.ReadFile(fileno(), 0, True)
                        except OSError as e:
                            err = e.winerror
                            if err not in _ready_errors:
                                raise
                        if err == _winapi.ERROR_IO_PENDING:
                            ov_list.append(ov)
                            waithandle_to_obj[ov.event] = o
                        else:
                            # If o.fileno() is an overlapped pipe handle and
                            # err == 0 then there is a zero length message
                            # in the pipe, but it HAS NOT been consumed.
                            ready_objects.add(o)
                            timeout = 0

                ready_handles = _exhaustive_wait(
                    waithandle_to_obj.keys(), timeout
                )
            finally:
                # request that overlapped reads stop
                for ov in ov_list:
                    ov.cancel()

                # wait for all overlapped reads to stop
                for ov in ov_list:
                    try:
                        _, err = ov.GetOverlappedResult(True)
                    except OSError as e:
                        err = e.winerror
                        if err not in _ready_errors:
                            raise
                    if err != _winapi.ERROR_OPERATION_ABORTED:
                        o = waithandle_to_obj[ov.event]
                        ready_objects.add(o)
                        if err == 0:
                            # If o.fileno() is an overlapped pipe handle then
                            # a zero length message HAS been consumed.
                            if hasattr(o, '_got_empty_message'):
                                o._got_empty_message = True

            ready_objects.update(waithandle_to_obj[h] for h in ready_handles)
            return [p for p in object_list if p in ready_objects]

    else:
        if hasattr(select, 'poll'):

            def _poll(fds, timeout):
                if timeout is not None:
                    timeout = int(timeout * 1000)  # timeout is in milliseconds
                fd_map = {}
                pollster = select.poll()
                for fd in fds:
                    pollster.register(fd, select.POLLIN)
                    if hasattr(fd, 'fileno'):
                        fd_map[fd.fileno()] = fd
                    else:
                        fd_map[fd] = fd
                ls = []
                for fd, event in pollster.poll(timeout):
                    if event & select.POLLNVAL:
                        raise ValueError('invalid file descriptor %i' % fd)
                    ls.append(fd_map[fd])
                return ls

        else:

            def _poll(fds, timeout):
                return select.select(fds, [], [], timeout)[0]

        def wait(object_list, timeout=None):
            '''
            Wait till an object in object_list is ready/readable.
            Returns list of those objects in object_list which are
             ready/readable.
            '''
            if timeout is not None:
                if timeout <= 0:
                    return _poll(object_list, 0)
                else:
                    deadline = time.time() + timeout
            while True:
                try:
                    return _poll(object_list, timeout)
                except OSError as e:
                    if e.errno != errno.EINTR:
                        raise
                if timeout is not None:
                    timeout = deadline - time.time()

    class Socket(object):
        """A Socket compatible with multiprocessing.connection.Client, that
        uses socket objects."""

        # https://github.com/akheron/cpython/blob/3.3/Lib/multiprocessing/connection.py#L349
        def __init__(self, address):
            self._handle = socket.socket()
            self._handle.connect(address)
            self._handle.setblocking(1)

        def send_bytes(self, buf):
            self._check_closed()
            n = len(buf)
            # For wire compatibility with 3.2 and lower
            header = struct.pack("!i", n)
            if n > 16384:
                # The payload is large so Nagle's algorithm won't be triggered
                # and we'd better avoid the cost of concatenation.
                chunks = [header, buf]
            elif n > 0:
                # Issue #20540: concatenate before sending, to avoid delays
                # dueto Nagle's algorithm on a TCP socket.
                chunks = [header + buf]
            else:
                # This code path is necessary to avoid "broken pipe" errors
                # when sending a 0-length buffer if the other end closed the
                # pipe.
                chunks = [header]
            for chunk in chunks:
                self._handle.sendall(chunk)

        def _safe_recv(self, *args, **kwargs):
            while True:
                try:
                    return self._handle.recv(*args, **kwargs)
                except socket.error as e:
                    # Interrupted system call
                    if e.errno != errno.EINTR:
                        raise

        def recv_bytes(self):
            self._check_closed()
            size, = struct.unpack("!i", self._safe_recv(4))
            return self._safe_recv(size)

        def _check_closed(self):
            if self._handle is None:
                raise IOError("handle is closed")

        def close(self):
            self._check_closed()
            self._handle.close()
            self._handle = None

        def poll(self, timeout=0.0):
            """Whether there is any input available to be read"""
            self._check_closed()
            return self._poll(timeout)

        def _poll(self, timeout):
            r = wait([self._handle], timeout)
            return bool(r)


else:
    from multiprocessing.connection import Client as Socket

try:
    from importlib.util import find_spec
    from importlib import import_module

    def existing_module(module):
        return bool(find_spec(module))


except ImportError:
    import imp

    def existing_module(module):
        try:
            imp.find_module(module)
            return True
        except ImportError:
            return False

    def import_module(module):
        __import__(module)
        if module not in sys.modules:
            raise ImportError(module)
        return sys.modules[module]


# Not really compat but convenient
try:
    from log_colorizer import get_color_logger
except ImportError:
    import logging

    logger = logging.getLogger
else:
    logger = get_color_logger



# *-* coding: utf-8 *-*
# This file is part of wdb
#
# wdb Copyright (c) 2012-2016  Florian Mounier, Kozea
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from . import (
    trace,
    start_trace,
    stop_trace,
    set_trace,
    Wdb,
    WEB_SERVER,
    WEB_PORT,
)
from .ui import dump
from ._compat import to_bytes, escape, logger, TCPServer

import traceback
from threading import current_thread
from uuid import uuid4
import sys

log = logger(__name__)
_exc_cache = {}


def _patch_tcpserver():
    """
    Patch shutdown_request to open blocking interaction after the end of the
    request
    """
    shutdown_request = TCPServer.shutdown_request

    def shutdown_request_patched(*args, **kwargs):
        thread = current_thread()
        shutdown_request(*args, **kwargs)
        if thread in _exc_cache:
            post_mortem_interaction(*_exc_cache.pop(thread))

    TCPServer.shutdown_request = shutdown_request_patched


def post_mortem_interaction(uuid, exc_info):
    wdb = Wdb.get(force_uuid=uuid)
    type_, value, tb = exc_info
    frame = None
    _value = value
    if not isinstance(_value, BaseException):
        _value = type_(value)

    wdb.obj_cache[id(exc_info)] = exc_info
    wdb.extra_vars['__exception__'] = exc_info
    exception = type_.__name__
    exception_description = str(value) + ' [POST MORTEM]'
    init = 'Echo|%s' % dump(
        {
            'for': '__exception__',
            'val': escape('%s: %s') % (exception, exception_description),
        }
    )

    wdb.interaction(
        frame,
        tb,
        exception,
        exception_description,
        init=init,
        iframe_mode=True,
        timeout=3,
    )


def _handle_off(silent=False):
    if not silent:
        log.exception('Exception with wdb off')

    uuid = str(uuid4())
    _exc_cache[current_thread()] = (uuid, sys.exc_info())

    web_url = 'http://%s:%d/pm/session/%s' % (
        WEB_SERVER or 'localhost',
        WEB_PORT or 1984,
        uuid,
    )
    return to_bytes(
        '''<!DOCTYPE html>
        <html>
            <head>
                <title>WDB Post Mortem</title>
                <!--
                    %s
                -->
                <style>
                  html, body, iframe {
                    margin: 0;
                    padding: 0;
                    width: 100%%;
                    height: 100%%;
                    border: none;
                    overflow: hidden;
                    display: block;
                  }
                </style>
                <script>
                    addEventListener("message", function (e) {
                        if (e.data == 'activate') {
                            var request = new XMLHttpRequest();
                            request.open('GET', '/__wdb/on', true);
                            request.onload = function() {
                                location.reload(true);
                            }
                            request.send();
                        }
                    }, false);
                </script>
            </head>
            <body>
                <iframe
                    src="%s"
                    id="wdbframe">
                </iframe>
            </body>
        </html>
    '''
        % (traceback.format_exc(), web_url)
    )


class WdbMiddleware(object):
    def __init__(self, app, start_disabled=False):
        _patch_tcpserver()
        self.app = app
        Wdb.enabled = not start_disabled

    def __call__(self, environ, start_response):
        path = environ.get('PATH_INFO', '')
        if path == '/__wdb/on':
            # Enable wdb
            Wdb.enabled = True
            start_response('200 OK', [('Content-Type', 'text/html')])
            return (to_bytes('Wdb is now on'),)

        if path == '/__wdb/shell':

            def f():
                # Enable wdb
                wdb = Wdb.get()
                Wdb.enabled = True
                start_response(
                    '200 OK',
                    [('Content-Type', 'text/html'), ('X-Thing', wdb.uuid)],
                )
                yield to_bytes(' ' * 4096)
                wdb = set_trace()
                wdb.die()
                yield to_bytes('Exited')

            return f()
        if Wdb.enabled:

            def trace_wsgi(environ, start_response):
                wdb = Wdb.get()
                wdb.closed = False
                appiter = None
                try:
                    with trace(close_on_exit=True, under=self.app):
                        appiter = self.app(environ, start_response)
                        for item in appiter:
                            yield item
                except Exception:
                    exc_info = sys.exc_info()
                    try:
                        start_response(
                            '500 INTERNAL SERVER ERROR',
                            [('Content-Type', 'text/html')],
                        )
                    except AssertionError:
                        log.exception(
                            'Exception with wdb off and headers already set',
                            exc_info=exc_info,
                        )
                        yield '\n'.join(
                            traceback.format_exception(*exc_info)
                        ).replace('\n', '\n<br>\n').encode('utf-8')
                    else:
                        yield _handle_off()
                finally:
                    hasattr(appiter, 'close') and appiter.close()
                wdb.closed = False

            return trace_wsgi(environ, start_response)

        def catch(environ, start_response):
            appiter = None

            try:
                appiter = self.app(environ, start_response)
                for item in appiter:
                    yield item
            except Exception:
                exc_info = sys.exc_info()
                try:
                    start_response(
                        '500 INTERNAL SERVER ERROR',
                        [('Content-Type', 'text/html')],
                    )
                except AssertionError:
                    log.exception(
                        'Exception with wdb off and headers already set',
                        exc_info=exc_info,
                    )
                    yield '\n'.join(
                        traceback.format_exception(*exc_info)
                    ).replace('\n', '\n<br>\n').encode('utf-8')
                else:
                    yield _handle_off()
            finally:
                # Close set_trace debuggers
                stop_trace(close_on_exit=True)
                hasattr(appiter, 'close') and appiter.close()

        return catch(environ, start_response)


def wdb_tornado(application, start_disabled=False):
    from tornado.web import (
        RequestHandler,
        ErrorHandler,
        HTTPError,
        StaticFileHandler,
    )
    from tornado.gen import coroutine

    Wdb.enabled = not start_disabled

    class WdbOn(RequestHandler):
        def get(self):
            Wdb.enabled = True
            self.write('Wdb is now on')

    class WdbOff(RequestHandler):
        def get(self):
            Wdb.enabled = False
            self.write('Wdb is now off')

    application.add_handlers(
        r'.*', ((r'/__wdb/on', WdbOn), (r'/__wdb/off', WdbOff))
    )
    old_execute = RequestHandler._execute
    under = getattr(RequestHandler._execute, '__wrapped__', None)

    @coroutine
    def _wdb_execute(*args, **kwargs):
        from wdb import trace, Wdb

        if Wdb.enabled:
            wdb = Wdb.get()
            wdb.closed = False  # Activate request ignores

        interesting = True
        if len(args) > 0 and isinstance(args[0], ErrorHandler):
            interesting = False
        elif (
            len(args) > 2
            and isinstance(args[0], StaticFileHandler)
            and args[2] == 'favicon.ico'
        ):
            interesting = False

        if Wdb.enabled and interesting:
            with trace(close_on_exit=True, under=under):
                old_execute(*args, **kwargs)
        else:
            old_execute(*args, **kwargs)
            # Close set_trace debuggers
            stop_trace(close_on_exit=True)

        if Wdb.enabled:
            # Reset closed state
            wdb.closed = False

    RequestHandler._execute = _wdb_execute

    def _wdb_error_writter(self, status_code, **kwargs):
        silent = False
        ex = kwargs.get('exc_info')
        if ex:
            silent = issubclass(ex[0], HTTPError)
        self.finish(_handle_off(silent=silent))
        post_mortem_interaction(*_exc_cache.pop(current_thread()))

    RequestHandler.write_error = _wdb_error_writter


def add_w_builtin():
    class w(object):
        """Global shortcuts"""

        @property
        def tf(self):
            set_trace(sys._getframe().f_back)

        @property
        def start(self):
            start_trace(sys._getframe().f_back)

        @property
        def stop(self):
            stop_trace(sys._getframe().f_back)

        @property
        def trace(self):
            trace(sys._getframe().f_back)

    __builtins__['w'] = w()


def patch_werkzeug():
    """Replace werkzeug debug middleware"""
    try:
        from werkzeug import debug
    except ImportError:
        return
    debug.DebuggedApplication = WdbMiddleware



import os.path
from hashlib import sha1
from ._compat import import_module, logger

log = logger('wdb.bp')


def canonic(filename):
    if filename == "<" + filename[1:-1] + ">":
        return filename
    canonic = os.path.abspath(filename)
    canonic = os.path.normcase(canonic)
    if canonic.endswith(('.pyc', '.pyo')):
        canonic = canonic[:-1]
    return canonic


def file_from_import(filename, function=None):
    try:
        module = import_module(filename)
    except ImportError:
        return filename
    if function is None:
        return module.__file__
    fun = getattr(module, function, None)
    if not fun or not hasattr(fun, '__code__'):
        return filename
    return fun.__code__.co_filename


class Breakpoint(object):
    """Simple breakpoint that breaks if in file"""

    def __init__(self, file, temporary=False):
        self.fn = file
        if not file.endswith(('.py', '.pyc', '.pyo')):
            file = file_from_import(file)
        self.file = canonic(file)
        self.temporary = temporary

    def on_file(self, filename):
        return canonic(filename) == self.file

    def breaks(self, frame):
        return self.on_file(frame.f_code.co_filename)

    def __repr__(self):
        s = 'Temporary ' if self.temporary else ''
        s += self.__class__.__name__
        s += ' on file %s' % self.file
        return s

    def __eq__(self, other):
        return self.file == other.file and self.temporary == other.temporary

    def __hash__(self):
        s = sha1()
        s.update(repr(self).encode('utf-8'))
        return int(s.hexdigest(), 16)

    def to_dict(self):
        return {
            'fn': self.file,
            'lno': getattr(self, 'line', None),
            'cond': getattr(self, 'condition', None),
            'fun': getattr(self, 'function', None),
            'temporary': self.temporary,
        }


class LineBreakpoint(Breakpoint):
    """Simple breakpoint that breaks if in file at line"""

    def __init__(self, file, line, temporary=False):
        self.line = line
        super(LineBreakpoint, self).__init__(file, temporary)

    def breaks(self, frame):
        return (
            super(LineBreakpoint, self).breaks(frame)
            and frame.f_lineno == self.line
        )

    def __repr__(self):
        return (
            super(LineBreakpoint, self).__repr__() + ' on line %d' % self.line
        )

    def __eq__(self, other):
        return (
            super(LineBreakpoint, self).__eq__(other)
            and self.line == other.line
        )

    def __hash__(self):
        return super(LineBreakpoint, self).__hash__()


class ConditionalBreakpoint(Breakpoint):
    """Breakpoint that breaks if condition is True at line in file"""

    def __init__(self, file, line, condition, temporary=False):
        self.line = line
        self.condition = condition
        super(ConditionalBreakpoint, self).__init__(file, temporary)

    def breaks(self, frame):
        try:
            return (
                super(ConditionalBreakpoint, self).breaks(frame)
                and (self.line is None or frame.f_lineno == self.line)
                and eval(self.condition, frame.f_globals, frame.f_locals)
            )
        except Exception:
            # Break in case of
            log.warning('Error in conditional break', exc_info=True)
            return True

    def __repr__(self):
        return (
            super(ConditionalBreakpoint, self).__repr__()
            + ' under the condition %s' % self.condition
        )

    def __eq__(self, other):
        return (
            super(ConditionalBreakpoint, self).__eq__(other)
            and self.condition == other.condition
        )

    def __hash__(self):
        return super(ConditionalBreakpoint, self).__hash__()


class FunctionBreakpoint(Breakpoint):
    """Breakpoint that breaks if in file in function"""

    def __init__(self, file, function, temporary=False):
        self.function = function
        if not file.endswith(('.py', '.pyc', '.pyo')):
            file = file_from_import(file, function)
        self.file = canonic(file)
        self.temporary = temporary

    def breaks(self, frame):
        return (
            super(FunctionBreakpoint, self).breaks(frame)
            and frame.f_code.co_name == self.function
        )

    def __repr__(self):
        return (
            super(FunctionBreakpoint, self).__repr__()
            + ' in function %s' % self.function
        )

    def __eq__(self, other):
        return (
            super(FunctionBreakpoint, self).__eq__(other)
            and self.function == other.function
        )

    def __hash__(self):
        return super(FunctionBreakpoint, self).__hash__()



import dis
import inspect
import io
import os
import signal
import sys
from contextlib import contextmanager
from difflib import HtmlDiff, _mdiff

from ._compat import OrderedDict, StringIO, existing_module


def pretty_frame(frame):
    if frame:
        return '%s <%s:%d>' % (
            frame.f_code.co_name,
            frame.f_code.co_filename,
            frame.f_lineno,
        )
    else:
        return 'None'


def get_code(obj):
    if hasattr(obj, '__func__'):
        return obj.__func__
    if hasattr(obj, '__code__'):
        return obj.__code__
    if hasattr(obj, 'gi_code'):
        return obj.gi_code
    if hasattr(obj, 'co_code'):
        return obj


def get_source_from_byte_code(code):
    try:
        import uncompyle6
    except Exception:
        return
    version = sys.version_info[0] + (sys.version_info[1] / 10.0)
    try:
        with open(os.devnull, 'w') as dn:
            return uncompyle6.deparse_code(version, code, dn).text
    except Exception:
        return


def get_source(obj):
    try:
        return inspect.getsource(obj)
    except Exception:
        code = get_code(obj)
        if code:
            source = get_source_from_byte_code(code)
            if source:
                return (
                    '# The following source has been decompilated:\n' + source
                )

        old_stdout = sys.stdout
        sys.stdout = StringIO()
        try:
            dis.dis(obj)
            sys.stdout.seek(0)
            rv = sys.stdout.read()
            sys.stdout = old_stdout
            return rv
        except Exception:
            sys.stdout = old_stdout
            return ''


def get_doc(obj):
    doc = inspect.getdoc(obj)
    com = inspect.getcomments(obj)
    if doc and com:
        return '%s\n\n(%s)' % (doc, com)
    elif doc:
        return doc
    elif com:
        return com
    return ''


def executable_line(line):
    line = line.strip()
    return not (
        (
            not line
            or (line[0] == '#')
            or (line[:3] == '"""')
            or line[:3] == "'''"
        )
    )


def get_args(frame):
    code = frame.f_code
    varnames = code.co_varnames
    nargs = code.co_argcount
    if hasattr(code, 'co_kwonlyargcount'):
        kwonly = code.co_kwonlyargcount
    else:
        # Python 2
        kwonly = 0
    locals = frame.f_locals

    # Regular args
    vars = OrderedDict([(var, locals[var]) for var in varnames[:nargs]])

    # Var args (*args)
    if frame.f_code.co_flags & 0x4:
        vars['*' + varnames[nargs + kwonly]] = locals[varnames[nargs + kwonly]]
        nargs += 1

    for n in range(kwonly):
        vars[varnames[nargs + n - 1]] = locals[varnames[nargs + n - 1]]

    nargs += kwonly

    if frame.f_code.co_flags & 0x8:
        vars['**' + varnames[nargs]] = locals[varnames[nargs]]

    return vars


def importable_module(module):
    return existing_module(module)


class Html5Diff(HtmlDiff):
    _table_template = """
      <table class="diff">
        %(header_row)s
        <tbody>
          %(data_rows)s
        </tbody>
      </table>"""

    def _format_line(self, side, flag, linenum, text):
        """Returns HTML markup of "from" / "to" text lines

        side -- 0 or 1 indicating "from" or "to" text
        flag -- indicates if difference on line
        linenum -- line number (used for line number column)
        text -- line text to be marked up
        """
        try:
            linenum = '%d' % linenum
            id = ' id="%s%s"' % (self._prefix[side], linenum)
        except TypeError:
            # handle blank lines where linenum is '>' or ''
            id = ''
        # replace those things that would get confused with HTML symbols
        text = (
            text.replace("&", "&amp;")
            .replace(">", "&gt;")
            .replace("<", "&lt;")
        )

        type_ = 'neutral'
        if '\0+' in text:
            type_ = 'add'
        if '\0-' in text:
            if type_ == 'add':
                type_ = 'chg'
            type_ = 'sub'
        if '\0^' in text:
            type_ = 'chg'

        # make space non-breakable so they don't get compressed or line wrapped
        text = text.replace(' ', '&nbsp;').rstrip()

        return (
            '<td class="diff_lno"%s>%s</td>'
            '<td class="diff_line diff_line_%s">%s</td>'
            % (id, linenum, type_, text)
        )

    def make_table(
        self,
        fromlines,
        tolines,
        fromdesc='',
        todesc='',
        context=False,
        numlines=5,
    ):
        """Returns HTML table of side by side comparison with change highlights

        Arguments:
        fromlines -- list of "from" lines
        tolines -- list of "to" lines
        fromdesc -- "from" file column header string
        todesc -- "to" file column header string
        context -- set to True for contextual differences (defaults to False
            which shows full differences).
        numlines -- number of context lines.  When context is set True,
            controls number of lines displayed before and after the change.
            When context is False, controls the number of lines to place
            the "next" link anchors before the next change (so click of
            "next" link jumps to just before the change).
        """

        # make unique anchor prefixes so that multiple tables may exist
        # on the same page without conflict.
        self._make_prefix()

        # change tabs to spaces before it gets more difficult after we insert
        # markup
        fromlines, tolines = self._tab_newline_replace(fromlines, tolines)

        # create diffs iterator which generates side by side from/to data
        if context:
            context_lines = numlines
        else:
            context_lines = None
            diffs = _mdiff(
                fromlines,
                tolines,
                context_lines,
                linejunk=self._linejunk,
                charjunk=self._charjunk,
            )

        # set up iterator to wrap lines that exceed desired width
        if self._wrapcolumn:
            diffs = self._line_wrapper(diffs)

        # collect up from/to lines and flags into lists (also format the lines)
        fromlist, tolist, flaglist = self._collect_lines(diffs)

        # process change flags, generating middle column of next anchors/links
        fromlist, tolist, flaglist, next_href, next_id = self._convert_flags(
            fromlist, tolist, flaglist, context, numlines
        )

        s = []
        fmt = ' <tr>%s%s</tr>\n'

        for i in range(len(flaglist)):
            if flaglist[i] is None:
                # mdiff yields None on separator lines skip the bogus ones
                # generated for the first line
                if i > 0:
                    s.append('        </tbody>        \n        <tbody>\n')
            else:
                s.append(fmt % (fromlist[i], tolist[i]))
        if fromdesc or todesc:
            header_row = '<thead><tr>%s%s</tr></thead>' % (
                '<th colspan="2" class="diff_header">%s</th>' % fromdesc,
                '<th colspan="2" class="diff_header">%s</th>' % todesc,
            )
        else:
            header_row = ''

        table = self._table_template % dict(
            data_rows=''.join(s), header_row=header_row, prefix=self._prefix[1]
        )

        return (
            table.replace('\0+', '<span class="diff_add">')
            .replace('\0-', '<span class="diff_sub">')
            .replace('\0^', '<span class="diff_chg">')
            .replace('\1', '</span>')
            .replace('\t', '&nbsp;')
        )


def search_key_in_obj(key, obj, matches=None, path='', context=None):
    context = context or []
    matches = matches or []
    if id(obj) in context:
        return matches
    context.append(id(obj))

    if isinstance(obj, dict):
        for k, v in obj.items():
            if not isinstance(k, str):
                continue
            if isinstance(v, type(sys)):
                continue
            if key in k:
                matches.append(
                    (
                        "%s['%s']"
                        % (
                            path.rstrip('.'),
                            k.replace(key, '<mark>%s</mark>' % key),
                        ),
                        v,
                    )
                )
            try:
                matches = search_key_in_obj(
                    key,
                    v,
                    matches,
                    "%s['%s']." % (path.rstrip('.'), k),
                    context,
                )
            except Exception:
                pass

    if isinstance(obj, list):
        for i, v in enumerate(obj):
            if isinstance(v, type(sys)):
                continue
            try:
                matches = search_key_in_obj(
                    key, v, matches, "%s[%d]." % (path.rstrip('.'), i), context
                )
            except Exception:
                pass

    for k in dir(obj):
        if k.startswith('__') and k not in ('__class__',):
            continue
        v = getattr(obj, k, None)
        v2 = getattr(obj, k, None)
        if id(v) != id(v2):
            continue
        if isinstance(v, type(sys)):
            continue
        if key in k:
            matches.append(
                ('%s%s' % (path, k.replace(key, '<mark>%s</mark>' % key)), v)
            )
        try:
            matches = search_key_in_obj(
                key, v, matches, '%s%s.' % (path, k), context
            )
        except Exception:
            pass

    return matches


def search_value_in_obj(fun, obj, matches=None, path='', context=None):
    context = context or []
    matches = matches or []
    if id(obj) in context:
        return matches
    context.append(id(obj))

    if isinstance(obj, dict):
        for k, v in obj.items():
            if not isinstance(k, str):
                continue
            if isinstance(v, type(sys)):
                continue

            res = None
            try:
                res = eval(fun, {'x': v})
            except Exception:
                pass
            new_path = "%s['%s']" % (path.rstrip('.'), k)

            if res:
                matches.append((new_path, v))
            try:
                matches = search_value_in_obj(
                    fun, v, matches, new_path + '.', context
                )
            except Exception:
                pass

    if isinstance(obj, list):
        for i, v in enumerate(obj):
            if isinstance(v, type(sys)):
                continue

            res = None
            try:
                res = eval(fun, {'x': v})
            except Exception:
                pass

            new_path = "%s[%d]" % (path.rstrip('.'), i)

            if res:
                matches.append((new_path, v))
            try:
                matches = search_value_in_obj(
                    fun, v, matches, new_path + '.', context
                )
            except Exception:
                pass

    for k in dir(obj):
        if k.startswith('__') and k not in ('__class__',):
            continue
        v = getattr(obj, k, None)
        v2 = getattr(obj, k, None)
        if id(v) != id(v2):
            continue
        if isinstance(v, type(sys)):
            continue

        res = None
        try:
            res = eval(fun, {'x': v})
        except Exception:
            pass

        new_path = '%s%s' % (path, k)

        if res:
            matches.append((new_path, v))
        try:
            matches = search_value_in_obj(
                fun, v, matches, new_path + '.', context
            )
        except Exception:
            pass

    return matches


class timeout_of(object):
    def __init__(self, time, strict=False):
        self.time = time
        try:
            # Ignoring when not active + disabling if no alarm signal (Windows)
            signal.signal(signal.SIGALRM, signal.SIG_IGN)
        except Exception:
            if strict:
                raise Exception('Not running because timeout is not available')
            self.active = False
        else:
            self.active = True

    def timeout(self, signum, frame):
        raise Exception('Timeout')

    def __enter__(self):
        if not self.active:
            return

        signal.signal(signal.SIGALRM, self.timeout)
        signal.setitimer(signal.ITIMER_REAL, self.time)

    def __exit__(self, *args):
        if not self.active:
            return

        signal.setitimer(signal.ITIMER_REAL, 0)
        signal.signal(signal.SIGALRM, signal.SIG_IGN)


class IterableEllipsis(object):
    def __init__(self, size):
        self.size = size


def cut_if_too_long(iterable, level, tuple_=False):
    max_ = 100
    for i in range(1, min(level, 4)):
        max_ /= 2
    start = 10
    end = 5
    max_ = max(start + end, int(max_))
    size = len(iterable)
    if size > max_:
        ie = IterableEllipsis(size - start - end)
        if tuple_:
            ie = (ie, ie)
        return list(iterable[:start]) + [ie] + list(iterable[-end:])
    else:
        return iterable


# Got from:
# http://www.zopatista.com/python/2013/11/26/inplace-file-rewriting/
# as suggested by https://github.com/leorochael
@contextmanager
def inplace(
    filename,
    mode='r',
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    backup_extension=None,
):
    """Allow for a file to be replaced with new content.

    yields a tuple of (readable, writable) file objects, where writable
    replaces readable.

    If an exception occurs, the old file is restored, removing the
    written data.

    mode should *not* use 'w', 'a' or '+'; only read-only-modes are supported.

    """

    # move existing file to backup, create new file with same permissions
    # borrowed extensively from the fileinput module
    if set(mode).intersection('wa+'):
        raise ValueError('Only read-only file modes can be used')

    backupfilename = filename + (backup_extension or os.extsep + 'bak')
    try:
        os.unlink(backupfilename)
    except os.error:
        pass
    os.rename(filename, backupfilename)
    readable = io.open(
        backupfilename,
        mode,
        buffering=buffering,
        encoding=encoding,
        errors=errors,
        newline=newline,
    )
    try:
        perm = os.fstat(readable.fileno()).st_mode
    except OSError:
        writable = io.open(
            filename,
            'w' + mode.replace('r', ''),
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )
    else:
        os_mode = os.O_CREAT | os.O_WRONLY | os.O_TRUNC
        if hasattr(os, 'O_BINARY'):
            os_mode |= os.O_BINARY
        fd = os.open(filename, os_mode, perm)
        writable = io.open(
            fd,
            "w" + mode.replace('r', ''),
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
        )
        try:
            if hasattr(os, 'chmod'):
                os.chmod(filename, perm)
        except OSError:
            pass
    try:
        yield readable, writable
    except Exception:
        # move backup back
        try:
            os.unlink(filename)
        except os.error:
            pass
        os.rename(backupfilename, filename)
        raise
    finally:
        readable.close()
        writable.close()
        try:
            os.unlink(backupfilename)
        except os.error:
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
qwdb
"""
import sys

from setuptools import setup

__version__ = '3.3.2'

requires = [
    "log_colorizer>=1.8.3",
    "jedi>=0.9.0",
    'uncompyle6',
    'python-magic>=0.4.15',
]

if sys.version_info[:2] <= (2, 6):
    requires.append('argparse')
    requires.append('ordereddict')
else:
    requires.append('importmagic3')

options = dict(
    name="qwdb",
    version=__version__,
    description="An improbable web debugger through WebSockets (client only)",
    long_description="See http://github.com/quilsoft-org/wdb",
    author="Florian Mounier @ kozea",
    author_email="florian.mounier@kozea.fr",
    url="http://github.com/quilsoft-org/wdb",
    license="GPLv3",
    platforms="Any",
    packages=['wdb'],
    install_requires=requires,
    entry_points={
        'console_scripts': [
            'wdb=wdb.__main__:main',
            'wdb-%s=wdb.__main__:main' % sys.version[:3],
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Debuggers",
    ],
)

setup(**options)
